@@ -16,7 +16,6 @@ from pycognito import Cognito
 
 from .. import (
     _CREATE_AUDIT_RECORDS,
-    _GET_AWS_CREDENTIALS_GQL,
     _GET_BULK_DATA_STORAGE_GQL,
     _GET_NODE_GQL,
 )
@@ -40,40 +39,47 @@ class _AuditRecordQueue(asyncio.Queue):
         super().__init__()
 
         async def sender() -> None:
+            cancelled = asyncio.Event()
+
             async def batcher() -> AsyncGenerator[
                 list[dict],
                 None,
             ]:
                 batch: list[dict] = list()
-                while True:
+                while not (cancelled.is_set() and self.empty()):
                     try:
-                        batch.append(
-                            await asyncio.wait_for(self.get(), timeout=node.timeout)
-                        )
-                    except asyncio.TimeoutError:
+                        try:
+                            batch.append(
+                                await asyncio.wait_for(self.get(), timeout=node.timeout)
+                            )
+                        except asyncio.TimeoutError:
+                            if batch:
+                                yield batch
+                                batch = list()
+                        else:
+                            if len(batch) == 500:
+                                yield batch
+                                batch = list()
+                    except asyncio.CancelledError:
+                        cancelled.set()
                         if batch:
                             yield batch
                             batch = list()
-                    else:
-                        if len(batch) == 500:
-                            yield batch
-                            batch = list()
 
             async for batch in batcher():
                 try:
-                    async with node._gql_client_lock:
-                        async with node._gql_client as session:
-                            await session.execute(
-                                _CREATE_AUDIT_RECORDS,
-                                variable_values=dict(
-                                    name=node.name,
-                                    tenant=node.tenant,
-                                    messageType=message_type.name,
-                                    auditRecords=batch,
-                                ),
-                            )
+                    async with node._gql_client as session:
+                        await session.execute(
+                            _CREATE_AUDIT_RECORDS,
+                            variable_values=dict(
+                                name=node.name,
+                                tenant=node.tenant,
+                                messageType=message_type.name,
+                                auditRecords=batch,
+                            ),
+                        )
                 except asyncio.CancelledError:
-                    raise
+                    cancelled.set()
                 except Exception:
                     getLogger().exception("Error creating audit records")
                 finally:
@@ -86,47 +92,7 @@ class _AuditRecordQueue(asyncio.Queue):
         return await super().get()
 
 
-class _BulkDataStorageQueue(asyncio.Queue):
-    def __init__(self, node: Node) -> None:
-        super().__init__()
-        self.__client_session = ClientSession()
-
-        async def filler() -> None:
-            while True:
-                if self.qsize() >= 20:
-                    await asyncio.sleep(0)
-                    continue
-                try:
-                    async with node._gql_client_lock:
-                        async with node._gql_client as session:
-                            bulk_data_storages = await session.execute(
-                                _GET_BULK_DATA_STORAGE_GQL,
-                                variable_values={"tenant", node.tenant},
-                            )
-                except asyncio.CancelledError:
-                    raise
-                except Exception:
-                    getLogger().exception("Error getting bulk data storage")
-                else:
-                    for bulk_data_storage in bulk_data_storages:
-                        await self.put(
-                            BulkDataStorage(bulk_data_storage, self.__client_session)
-                        )
-
-        asyncio.create_task(filler(), name="BulkDataStorageQueueFiller")
-
-    def _get(self) -> BulkDataStorage:
-        bulk_data_storage: BulkDataStorage = super()._get()
-        return bulk_data_storage if not bulk_data_storage.expired else self._get()
-
-    async def get(self) -> BulkDataStorage:
-        return await super().get()
-
-    async def stop(self) -> None:
-        await self.__client_session.close()
-
-
-class BulkDataStorage(BaseBulkDataStorage):
+class _BulkDataStorage(BaseBulkDataStorage):
     def __init__(
         self,
         bulk_data_storage: dict[str, Union[str, PresignedPost]],
@@ -151,34 +117,91 @@ class BulkDataStorage(BaseBulkDataStorage):
         return self.presigned_get
 
 
+class _BulkDataStorageQueue(asyncio.Queue):
+    def __init__(self, node: Node) -> None:
+        super().__init__()
+        self.__fill = asyncio.Event()
+
+        async def filler() -> None:
+            async with ClientSession() as client_session:
+                cancelled = asyncio.Event()
+                while not cancelled.is_set():
+                    bulk_data_storages: list[dict] = list()
+                    try:
+                        await self.__fill.wait()
+                        async with node._gql_client as session:
+                            bulk_data_storages = (
+                                await session.execute(
+                                    _GET_BULK_DATA_STORAGE_GQL,
+                                    variable_values={"tenant": node.tenant},
+                                )
+                            )["GetBulkDataStorage"]
+                    except asyncio.CancelledError:
+                        cancelled.set()
+                    except Exception:
+                        getLogger().exception("Error getting bulk data storage")
+                    for bulk_data_storage in bulk_data_storages:
+                        self.put_nowait(
+                            _BulkDataStorage(bulk_data_storage, client_session)
+                        )
+                    self.__fill.clear()
+
+        asyncio.create_task(filler(), name="BulkDataStorageQueueFiller")
+
+    async def get(self) -> _BulkDataStorage:
+        if self.qsize() < 20:
+            self.__fill.set()
+        bulk_data_storage: _BulkDataStorage = await super().get()
+        return bulk_data_storage if not bulk_data_storage.expired else await self.get()
+
+
+class _CognitoAIOHTTPTransport(AIOHTTPTransport):
+    def __init__(self, cognito: Cognito, url: str, **kwargs: Any) -> None:
+        self._cognito = cognito
+        super().__init__(url, **kwargs)
+
+    def __getattribute__(self, name: str) -> Any:
+        if name == "headers":
+            self._cognito.check_token()
+            return dict(Authorization=self._cognito.access_token)
+        return super().__getattribute__(name)
+
+
 class _DeleteMessageQueue(asyncio.Queue):
     def __init__(self, edge: Edge, node: Node) -> None:
         super().__init__()
 
         async def deleter() -> None:
+            cancelled = asyncio.Event()
+
             async def batcher() -> AsyncGenerator[
                 list[str],
                 None,
             ]:
                 batch: list[str] = list()
-                while True:
+                while not (cancelled.is_set() and self.empty()):
                     try:
-                        batch.append(
-                            await asyncio.wait_for(self.get(), timeout=node.timeout)
-                        )
-                    except asyncio.TimeoutError:
+                        try:
+                            batch.append(
+                                await asyncio.wait_for(self.get(), timeout=node.timeout)
+                            )
+                        except asyncio.TimeoutError:
+                            if batch:
+                                yield batch
+                                batch = list()
+                        else:
+                            if len(batch) == 10:
+                                yield batch
+                                batch = list()
+                    except asyncio.CancelledError:
+                        cancelled.set()
                         if batch:
                             yield batch
                             batch = list()
-                    else:
-                        if len(batch) == 10:
-                            yield batch
-                            batch = list()
 
-            loop = asyncio.get_running_loop()
             async for receipt_handles in batcher():
                 try:
-                    response = await loop.run_in_executor(
+                    response = await asyncio.get_running_loop().run_in_executor(
                         None,
                         partial(
                             node._sqs_client.delete_message_batch,
@@ -191,16 +214,16 @@ class _DeleteMessageQueue(asyncio.Queue):
                             QueueUrl=edge.queue,
                         ),
                     )
+                except asyncio.CancelledError:
+                    cancelled.set()
+                except Exception:
+                    getLogger().exception(f"Error deleting messages from {edge.name}")
+                finally:
                     for failed in response.get("Failed", list()):
                         id = failed.pop("Id")
                         getLogger().error(
-                            f"Unable to send message {receipt_handles[id]} to {edge.name}, reason {failed}"
+                            f"Unable to delete message {receipt_handles[id]} from {edge.name}, reason {failed}"
                         )
-                except asyncio.CancelledError:
-                    raise
-                except Exception:
-                    getLogger().exception(f"Error sending messages to {edge.name}")
-                finally:
                     for _ in range(len(receipt_handles)):
                         self.task_done()
 
@@ -212,18 +235,14 @@ class _DeleteMessageQueue(asyncio.Queue):
 
 class _SourceMessageReceiver:
     def __init__(self, edge: Edge, node: Node) -> None:
-        self.__continue = asyncio.Event()
-        self.__continue.set()
         self.__delete_message_queue = _DeleteMessageQueue(edge, node)
 
         async def receive() -> None:
-            await self.__continue.wait()
-            loop = asyncio.get_running_loop()
             getLogger().info(f"Receiving messages from {edge.name}")
             error_count = 0
-            while self.__continue.is_set():
+            while True:
                 try:
-                    response = await loop.run_in_executor(
+                    response = await asyncio.get_running_loop().run_in_executor(
                         None,
                         partial(
                             node._sqs_client.receive_message,
@@ -253,10 +272,7 @@ class _SourceMessageReceiver:
                         )
                         await asyncio.sleep(10)
                 else:
-                    if not (
-                        self.__continue.is_set()
-                        and (sqs_messages := response.get("Messages"))
-                    ):
+                    if not (sqs_messages := response.get("Messages")):
                         continue
                     getLogger().info(f"Received {len(sqs_messages)} from {edge.name}")
                     for sqs_message in sqs_messages:
@@ -282,62 +298,73 @@ class _SourceMessageReceiver:
                             getLogger().exception(
                                 f"Error handling recevied message for {edge.name}"
                             )
-                        self.__delete_message_queue.put_nowait(receipt_handle)
+                        else:
+                            self.__delete_message_queue.put_nowait(receipt_handle)
             getLogger().info(f"Stopping receiving messages from {edge.name}")
 
-        asyncio.create_task(receive(), name=f"SourceMessageReceiver({edge.name})")
+        self.__task = asyncio.create_task(
+            receive(), name=f"SourceMessageReceiver({edge.name})"
+        )
 
     async def join(self) -> None:
-        done, pending = await asyncio.wait([self.__task])
+        await asyncio.wait([self.__task])
         await self.__delete_message_queue.join()
 
-    def stop(self) -> None:
-        self.__continue.clear()
-
 
 class _TargetMessageQueue(asyncio.Queue):
     def __init__(self, node: Node, edge: Edge) -> None:
+        super().__init__()
+
         async def sender() -> None:
+            cancelled = asyncio.Event()
+
             async def batcher() -> AsyncGenerator[
                 list[SendMessageBatchRequestEntryTypeDef], None
             ]:
                 batch: list[SendMessageBatchRequestEntryTypeDef] = list()
                 batch_length = 0
                 id = 0
-                while True:
+                while not (cancelled.is_set() and self.empty()):
                     try:
-                        message = await asyncio.wait_for(
-                            self.get(), timeout=node.timeout
-                        )
-                    except asyncio.TimeoutError:
+                        try:
+                            message = await asyncio.wait_for(
+                                self.get(), timeout=node.timeout
+                            )
+                        except asyncio.TimeoutError:
+                            if batch:
+                                yield batch
+                            batch = list()
+                            batch_length = 0
+                            id = 0
+                        else:
+                            if batch_length + len(message) > 262144:
+                                yield batch
+                                batch = list()
+                                batch_length = 0
+                                id = 0
+                            batch.append(
+                                SendMessageBatchRequestEntryTypeDef(
+                                    Id=str(id), **message._sqs_message(node)
+                                )
+                            )
+                            if len(batch) == 10:
+                                yield batch
+                                batch = list()
+                                batch_length = 0
+                                id = 0
+                            id += 1
+                            batch_length += len(message)
+                    except asyncio.CancelledError:
+                        cancelled.set()
                         if batch:
                             yield batch
                         batch = list()
                         batch_length = 0
                         id = 0
-                    else:
-                        if batch_length + len(message) > 262144:
-                            yield batch
-                            batch = list()
-                            batch_length = 0
-                            id = 0
-                        batch.append(
-                            SendMessageBatchRequestEntryTypeDef(
-                                Id=str(id), **message._sqs_message(node)
-                            )
-                        )
-                        if len(batch) == 10:
-                            yield batch
-                            batch = list()
-                            batch_length = 0
-                            id = 0
-                        id += 1
-                        batch_length += len(message)
 
-            loop = asyncio.get_running_loop()
             async for entries in batcher():
                 try:
-                    response = await loop.run_in_executor(
+                    response = await asyncio.get_running_loop().run_in_executor(
                         None,
                         partial(
                             node._sqs_client.send_message_batch,
@@ -345,16 +372,16 @@ class _TargetMessageQueue(asyncio.Queue):
                             QueueUrl=edge.queue,
                         ),
                     )
-                    for failed in response["Failed"]:
+                except asyncio.CancelledError:
+                    cancelled.set()
+                except Exception:
+                    getLogger().exception(f"Error sending messages to {edge.name}")
+                finally:
+                    for failed in response.get("Failed", list()):
                         id = failed.pop("Id")
                         getLogger().error(
                             f"Unable to send message {entries[id]} to {edge.name}, reason {failed}"
                         )
-                except asyncio.CancelledError:
-                    raise
-                except Exception:
-                    getLogger().exception(f"Error sending messages to {edge.name}")
-                finally:
                     for _ in range(len(entries)):
                         self.task_done()
 
@@ -364,18 +391,6 @@ class _TargetMessageQueue(asyncio.Queue):
         return await super().get()
 
 
-class CognitoAIOHTTPTransport(AIOHTTPTransport):
-    def __init__(self, cognito: Cognito, url: str, **kwargs: Any) -> None:
-        self._cognito = cognito
-        super().__init__(url, **kwargs)
-
-    def __getattribute__(self, name: str) -> Any:
-        if name == "headers":
-            self._cognito.check_token()
-            return dict(Authorization=self._cognito.access_token)
-        return super().__getattribute__(name)
-
-
 class Node(BaseNode):
     def __init__(
         self,
@@ -390,9 +405,9 @@ class Node(BaseNode):
         username: str = None,
     ) -> None:
         super().__init__(
-            gql_transport_cls=CognitoAIOHTTPTransport,
             appsync_endpoint=appsync_endpoint,
             client_id=client_id,
+            gql_transport_cls=_CognitoAIOHTTPTransport,
             name=name,
             password=password,
             tenant=tenant,
@@ -400,28 +415,11 @@ class Node(BaseNode):
             user_pool_id=user_pool_id,
             username=username,
         )
-        self._gql_client_lock = asyncio.Lock()
-        self.__bulk_data_storage_queue: _BulkDataStorageQueue = None
         self.__audit_records_queues: dict[str, _AuditRecordQueue] = dict()
+        self.__bulk_data_storage_queue: _BulkDataStorageQueue = None
         self.__source_message_receivers: list[_SourceMessageReceiver] = list()
-        self.__stop = asyncio.Event()
         self.__target_message_queues: dict[str, _TargetMessageQueue] = dict()
 
-    def _get_aws_credentials(self, duration: int = 3600) -> dict[str, str]:
-        loop = asyncio.get_event_loop()
-
-        async def get_aws_credentials() -> dict[str, str]:
-            async with self._gql_client_lock:
-                async with self._gql_client as session:
-                    return await session.execute(
-                        _GET_AWS_CREDENTIALS_GQL,
-                        variable_values=dict(
-                            name=self.app, tenant=self.tenant, duration=duration
-                        ),
-                    )["GetApp"]["GetAwsCredentials"]
-
-        return loop.run_until_complete(get_aws_credentials())
-
     def audit_message(
         self,
         /,
@@ -454,7 +452,6 @@ class Node(BaseNode):
         pass
 
     async def join(self) -> None:
-        await self.__stop.wait()
         await asyncio.gather(
             *[
                 source_message_receiver.join()
@@ -469,12 +466,8 @@ class Node(BaseNode):
         )
         for audit_records_queue in self.__audit_records_queues.values():
             await audit_records_queue.join()
-        if self.__bulk_data_storage_queue:
-            await self.__bulk_data_storage_queue.stop()
 
-    def send_message(
-        self, /, message: Message, *, targets: set[Edge] = None
-    ) -> None:
+    def send_message(self, /, message: Message, *, targets: set[Edge] = None) -> None:
         self.send_messages([message], targets=targets)
 
     def send_messages(
@@ -491,18 +484,19 @@ class Node(BaseNode):
                     getLogger().warning(f"Target {target.name} does not exist")
 
     async def start(self) -> None:
-        self.__stop.clear()
+        getLogger().info(f"Starting Node {self.name}")
         self.__bulk_data_storage_queue = _BulkDataStorageQueue(self)
-        async with self._gql_client_lock:
-            async with self._gql_client as session:
-                data: dict[str, dict] = await session.execute(
+        async with self._gql_client as session:
+            data: dict[str, Union[str, dict]] = (
+                await session.execute(
                     _GET_NODE_GQL,
                     variable_values=dict(name=self.name, tenant=self.tenant),
-                )["GetNode"]
+                )
+            )["GetNode"]
         self.config = (
-            json.loads(data["tenant"].get("config", {}))
-            | json.loads(data["app"].get("config", {}))
-            | json.loads(data.get("config", {}))
+            json.loads(data["tenant"].get("config") or "{}")
+            | json.loads((data.get("app") or dict()).get("config") or "{}")
+            | json.loads(data.get("config") or "{}")
         )
         if receive_message_type := data.get("receiveMessageType"):
             self._receive_message_type = MessageType(
@@ -518,15 +512,21 @@ class Node(BaseNode):
                 name=send_message_type["name"],
             )
             self.__audit_records_queues[send_message_type["name"]] = _AuditRecordQueue(
-                self.receive_message_type, self
+                self.send_message_type, self
             )
-        self._sources = {
-            Edge(name=edge["source"]["name"], queue=edge["queue"])
-            for edge in data["receiveEdges"]
-        }
+        if self.node_type == "AppChangeReceiverNode":
+            if edge := data.get("receiveEdge"):
+                self._sources = {Edge(name=edge["source"]["name"], queue=edge["queue"])}
+            else:
+                self._sources = set()
+        else:
+            self._sources = {
+                Edge(name=edge["source"]["name"], queue=edge["queue"])
+                for edge in (data.get("receiveEdges") or list())
+            }
         self._targets = {
             Edge(name=edge["target"]["name"], queue=edge["queue"])
-            for edge in data["sendEdges"]
+            for edge in (data.get("sendEdges") or list())
         }
         self.__target_message_queues = {
             edge.name: _TargetMessageQueue(self, edge) for edge in self._targets
@@ -535,7 +535,6 @@ class Node(BaseNode):
             _SourceMessageReceiver(edge, self) for edge in self._sources
         ]
 
-    async def stop(self) -> None:
-        for source_message_receiver in self.__source_message_receivers:
-            source_message_receiver.stop()
-        self.__stop.set()
+    async def start_and_run_forever(self) -> None:
+        await self.start()
+        await self.join()
