@@ -5,20 +5,17 @@ from datetime import datetime, timezone
 from gzip import GzipFile
 from io import BytesIO
 from queue import Empty, Queue
-from threading import Condition, Event, RLock, Thread
+from threading import Event, Thread
 from time import sleep
 from typing import TYPE_CHECKING, Any, BinaryIO, Generator, Union
 from uuid import uuid4
 
 import dynamic_function_loader
 from aws_error_utils import catch_aws_error
-from gql.transport.requests import RequestsHTTPTransport
-from pycognito import Cognito
 from requests import post
 
 from .. import (
     _CREATE_AUDIT_RECORDS,
-    _GET_AWS_CREDENTIALS_GQL,
     _GET_BULK_DATA_STORAGE_GQL,
     _GET_NODE_GQL,
 )
@@ -52,8 +49,8 @@ class _AuditRecordQueue(Queue):
                 if not batch:
                     continue
                 try:
-                    with node._gql_client_lock:
-                        node._gql_client.execute(
+                    with node._gql_client as session:
+                        session.execute(
                             _CREATE_AUDIT_RECORDS,
                             variable_values=dict(
                                 name=node.name,
@@ -93,32 +90,34 @@ class _BulkDataStorage(BaseBulkDataStorage):
 class _BulkDataStorageQueue(Queue):
     def __init__(self, node: Node) -> None:
         super().__init__()
-        self.fill = Condition()
+        self.__fill = Event()
 
         def filler() -> None:
             while True:
-                if not self.qsize() != 0:
-                    with self.fill:
-                        self.fill.wait()
-                with node._gql_client_lock:
-                    bulk_data_storages: list[dict] = node._gql_client.execute(
-                        _GET_BULK_DATA_STORAGE_GQL,
-                        variable_values={"tenant": node.tenant},
-                    )["GetBulkDataStorage"]
-                for bulk_data_storage in bulk_data_storages:
-                    self.put(_BulkDataStorage(bulk_data_storage))
+                self.__fill.wait()
+                try:
+                    with node._gql_client as session:
+                        bulk_data_storages: list[dict] = session.execute(
+                            _GET_BULK_DATA_STORAGE_GQL,
+                            variable_values={"tenant": node.tenant},
+                        )["GetBulkDataStorage"]
+                except Exception:
+                    getLogger().exception("Error getting bulk data storage")
+                else:
+                    for bulk_data_storage in bulk_data_storages:
+                        self.put_nowait(_BulkDataStorage(bulk_data_storage))
+                self.__fill.clear()
 
         Thread(daemon=True, name="BulkDataStorageQueueFiller", target=filler).start()
 
     def get(self, block: bool = True, timeout: float = None) -> _BulkDataStorage:
-        with self.fill:
-            if self.qsize() < 20:
-                self.fill.notify()
+        if self.qsize() < 20:
+            self.__fill.set()
         bulk_data_storage: _BulkDataStorage = super().get(block=block, timeout=timeout)
         return (
-            self.get(block=block, timeout=timeout)
-            if bulk_data_storage.expired
-            else bulk_data_storage
+            bulk_data_storage
+            if not bulk_data_storage.expired
+            else self.get(block=block, timeout=timeout)
         )
 
 
@@ -186,18 +185,6 @@ class _TargetMessageQueue(Queue):
         return super().get(block=block, timeout=timeout)
 
 
-class CognitoRequestsHTTPTransport(RequestsHTTPTransport):
-    def __init__(self, cognito: Cognito, url: str, **kwargs: Any) -> None:
-        self._cognito = cognito
-        super().__init__(url, **kwargs)
-
-    def __getattribute__(self, name: str) -> Any:
-        if name == "headers":
-            self._cognito.check_token()
-            return dict(Authorization=self._cognito.access_token)
-        return super().__getattribute__(name)
-
-
 class Node(BaseNode):
     def __init__(
         self,
@@ -212,7 +199,6 @@ class Node(BaseNode):
         username: str = None,
     ) -> None:
         super().__init__(
-            gql_transport_cls=CognitoRequestsHTTPTransport,
             appsync_endpoint=appsync_endpoint,
             client_id=client_id,
             name=name,
@@ -222,21 +208,11 @@ class Node(BaseNode):
             user_pool_id=user_pool_id,
             username=username,
         )
-        self._gql_client_lock = RLock()
         self.__bulk_data_storage_queue = _BulkDataStorageQueue(self)
         self.__audit_records_queues: dict[str, _AuditRecordQueue] = dict()
         self.__stop = Event()
         self.__target_message_queues: dict[str, _TargetMessageQueue] = dict()
 
-    def _get_aws_credentials(self, duration: int = 3600) -> dict[str, str]:
-        with self._gql_client_lock:
-            return self._gql_client.execute(
-                _GET_AWS_CREDENTIALS_GQL,
-                variable_values=dict(
-                    name=self.app, tenant=self.tenant, duration=duration
-                ),
-            )["GetApp"]["GetAwsCredentials"]
-
     def audit_message(
         self,
         /,
@@ -294,14 +270,14 @@ class Node(BaseNode):
     def start(self) -> None:
         getLogger().info(f"Starting Node {self.name}")
         self.__stop.clear()
-        with self._gql_client_lock:
-            data: dict[str, dict] = self._gql_client.execute(
+        with self._gql_client as session:
+            data: dict[str, Union[str, dict]] = session.execute(
                 _GET_NODE_GQL,
                 variable_values=dict(name=self.name, tenant=self.tenant),
             )["GetNode"]
         self.config = (
             json.loads(data["tenant"].get("config") or "{}")
-            | json.loads(data["app"].get("config") or "{}")
+            | json.loads((data.get("app") or dict()).get("config") or "{}")
             | json.loads(data.get("config") or "{}")
         )
         if receive_message_type := data.get("receiveMessageType"):
@@ -320,13 +296,19 @@ class Node(BaseNode):
             self.__audit_records_queues[send_message_type["name"]] = _AuditRecordQueue(
                 self._send_message_type, self
             )
-        self._sources = {
-            Edge(name=edge["source"]["name"], queue=edge["queue"])
-            for edge in data["receiveEdges"]
-        }
+        if self.node_type == "AppChangeReceiverNode":
+            if edge := data.get("receiveEdge"):
+                self._sources = {Edge(name=edge["source"]["name"], queue=edge["queue"])}
+            else:
+                self._sources = set()
+        else:
+            self._sources = {
+                Edge(name=edge["source"]["name"], queue=edge["queue"])
+                for edge in (data.get("receiveEdges") or list())
+            }
         self._targets = {
             Edge(name=edge["target"]["name"], queue=edge["queue"])
-            for edge in data["sendEdges"]
+            for edge in (data.get("sendEdges") or list())
         }
         self.__target_message_queues = {
             edge.name: _TargetMessageQueue(self, edge) for edge in self._targets
@@ -439,7 +421,8 @@ class _SourceMessageReceiver(Thread):
                             getLogger().exception(
                                 f"Error handling recevied message for {edge.name}"
                             )
-                        self.__delete_message_queue.put_nowait(receipt_handle)
+                        else:
+                            self.__delete_message_queue.put_nowait(receipt_handle)
             getLogger().info(f"Stopping receiving messages from {edge.name}")
 
         super().__init__(name=f"SourceMessageReceiver({edge.name})", target=receive)
@@ -489,6 +472,10 @@ class AppNode(Node):
             _SourceMessageReceiver(edge, self) for edge in self._sources
         ]
 
+    def start_and_run_forever(self) -> None:
+        self.start()
+        self.join()
+
     def stop(self) -> None:
         for app_node_receiver in self.__source_message_receivers:
             app_node_receiver.stop()
